@@ -28,6 +28,7 @@
 (require 'button)
 (require 'rx)
 
+(require 'persist)
 (require 'svg-lib)
 (require 'taxy)
 (require 'taxy-magit-section)
@@ -38,23 +39,64 @@
   "Group Ement rooms with Taxy."
   :group 'ement)
 
+;;;; Mouse commands
+
+;; Since mouse-activated commands must handle mouse events, we define a simple macro to
+;; wrap a command into a mouse-event-accepting one.
+
+(defmacro ement-room-list-define-mouse-command (command)
+  "Define a command that calls COMMAND interactively with point at mouse event.
+COMMAND should be a form that evaluates to a function symbol; if
+a symbol, it should be unquoted.."
+  (let ((docstring (format "Call command `%s' interactively with point at EVENT." command))
+        (name (intern (format "ement-room-list-mouse-%s" command))))
+    `(defun ,name (event)
+       ,docstring
+       (interactive "e")
+       (mouse-set-point event)
+       (call-interactively #',command))))
+
 ;;;; Variables
 
+(declare-function ement-room-toggle-space "ement-room")
+
 (defvar ement-room-list-mode-map
   (let ((map (make-sparse-keymap)))
     (define-key map (kbd "RET") #'ement-room-list-RET)
     (define-key map (kbd "SPC") #'ement-room-list-next-unread)
     (define-key map [tab] #'ement-room-list-section-toggle)
-    (define-key map [mouse-1] #'ement-room-list-mouse-1)
-    map))
+    (define-key map [mouse-1] (ement-room-list-define-mouse-command ement-room-list-RET))
+    (define-key map [mouse-2] (ement-room-list-define-mouse-command ement-room-list-kill-buffer))
+    (define-key map (kbd "k") #'ement-room-list-kill-buffer)
+    (define-key map (kbd "s") #'ement-room-toggle-space)
+    map)
+  "Keymap for `ement-room-list' buffers.
+See also `ement-room-list-button-map'.")
+
+(defvar ement-room-list-button-map
+  ;; This map is needed because some columns are propertized as buttons, which override
+  ;; the main keymap.
+  ;; TODO: Is it possible to adjust the button properties to obviate this map?
+  (let ((map (make-sparse-keymap)))
+    (define-key map [mouse-1] (ement-room-list-define-mouse-command ement-room-list-RET))
+    (define-key map [mouse-2] (ement-room-list-define-mouse-command ement-room-list-kill-buffer))
+    map)
+  "Keymap for buttonized text in `ement-room-list' buffers.")
 
 (defvar ement-room-list-timestamp-colors nil
   "List of colors used for timestamps.
 Set automatically when `ement-room-list-mode' is activated.")
 
+(defvar ement-room)
+(defvar ement-session)
 (defvar ement-sessions)
 (defvar ement-room-prism-minimum-contrast)
 
+;;;;; Persistent variables
+
+(persist-defvar ement-room-list-visibility-cache nil
+                "Applied to `magit-section-visibility-cache', which see.")
+
 ;;;; Customization
 
 (defcustom ement-room-list-auto-update t
@@ -263,18 +305,37 @@ from recent to non-recent for rooms updated in the past hour.")
       "Low-priority")))
 
 (defcustom ement-room-list-default-keys
-  '((space-p space)
+  '(;; First, group all invitations (this group will appear first since the rooms are
+    ;; already sorted first).
     ((membership :status 'invite))
-    (favourite)
-    (buffer)
+    ;; Group all left rooms (this group will appear last, because the rooms are already
+    ;; sorted last).
     ((membership :status 'leave))
-    (low-priority)
+    ;; Group all favorite rooms, which are already sorted first.
+    (favourite)
+    ;; Group all low-priority rooms, which are already sorted last, and within that group,
+    ;; group them by their space, if any.
+    (low-priority space)
+    ;; Group other rooms which are opened in a buffer.
+    (buffer)
+    ;; Group other rooms which are unread.
     (unread)
-    ((latest :name "Last 24h" :newer-than 86400))
-    (latest :name "Older than 90d" :older-than (* 86400 90))
-    people
-    freshness
-    (space))
+    ;; Group other rooms which are in a space by freshness, then by space.
+    ((and :name "Spaced"
+          :keys ((not space-p)
+                 space))
+     freshness space)
+    ;; Group spaces themselves by their parent space (since space headers can't also be
+    ;; items, we have to handle them separately; a bit of a hack, but not too bad).
+    ((and :name "Spaces" :keys (space-p))
+     space)
+    ;; Group rooms which aren't in spaces by their freshness.
+    ((and :name "Unspaced"
+          :keys ((not space)
+                 (not people)))
+     freshness)
+    ;; Group direct rooms by freshness.
+    (people freshness))
   "Default keys."
   :type 'sexp)
 
@@ -338,9 +399,10 @@ from recent to non-recent for rooms updated in the past hour.")
              (push 'ement-room-list-invited (map-elt face :inherit)))
             ('leave
              (push 'ement-room-list-left (map-elt face :inherit))))
-          (propertize (ement--button-buttonize display-name #'ement-room-list-mouse-1)
+          (propertize display-name
                       'face face
-                      'mouse-face 'highlight))
+                      'mouse-face 'highlight
+                      'keymap ement-room-list-button-map))
         "")))
 
 (ement-room-list-define-column #("Unread" 0 6 (help-echo "Unread events (Notifications:Highlights)")) (:align 'right)
@@ -455,8 +517,15 @@ from recent to non-recent for rooms updated in the past hour.")
   ;; visual bug that sometimes causes room avatars to be displayed in a section heading
   ;; when a section is hidden.
   (interactive)
-  (call-interactively #'magit-section-toggle)
-  (revert-buffer))
+  (ignore-errors
+    ;; Ignore an error in case point is past the top-level section.
+    (cl-typecase (aref (oref (magit-current-section) value) 0)
+      (ement-room
+       ;; HACK: Don't hide rooms themselves (they end up permanently hidden).
+       nil)
+      (otherwise
+       (call-interactively #'magit-section-toggle)
+       (revert-buffer)))))
 
 ;;;###autoload
 (defun ement-room-list--after-initial-sync (&rest _ignore)
@@ -470,15 +539,16 @@ To be called from `ement-after-initial-sync-hook'."
 ;;;###autoload
 (cl-defun ement-room-list (&key (buffer-name "*Ement Room List*")
                                 (keys ement-room-list-default-keys)
-                                (display-buffer-action '(display-buffer-same-window))
+                                (display-buffer-action '((display-buffer-reuse-window display-buffer-same-window)))
                                 ;; visibility-fn
                                 )
   "Show a buffer listing Ement rooms, grouped with Taxy KEYS.
-The buffer is named BUFFER-NAME and is shown with
-DISPLAY-BUFFER-ACTION; or if DISPLAY-BUFFER-ACTION is nil, the
-buffer is not displayed."
+After showing it, its window is selected.  The buffer is named
+BUFFER-NAME and is shown with DISPLAY-BUFFER-ACTION; or if
+DISPLAY-BUFFER-ACTION is nil, the buffer is not displayed."
   (interactive)
-  (let (format-table column-sizes window-start)
+  (let ((inhibit-read-only t)
+        format-table column-sizes window-start room-session-vectors)
     (cl-labels (;; (heading-face
                 ;;  (depth) (list :inherit (list 'bufler-group (bufler-level-face depth))))
                 (format-item (item) (gethash item format-table))
@@ -546,66 +616,73 @@ buffer is not displayed."
       ;;   (kill-buffer buffer-name))
       (unless ement-sessions
         (error "Ement: Not connected.  Use `ement-connect' to connect"))
+      (setf room-session-vectors
+            (cl-loop for (_id . session) in ement-sessions
+                     append (cl-loop for room in (ement-session-rooms session)
+                                     collect (vector room session))))
       (with-current-buffer (get-buffer-create buffer-name)
         (ement-room-list-mode)
-        (let* ((room-session-vectors
-                (cl-loop for (_id . session) in ement-sessions
-                         append (cl-loop for room in (ement-session-rooms session)
-                                         collect (vector room session))))
-               (taxy (cl-macrolet ((first-item
-                                    (pred) `(lambda (taxy)
-                                              (when (taxy-items taxy)
-                                                (,pred (car (taxy-items taxy)))))))
-                       (thread-last
-                         (make-fn
-                          :name "Ement Rooms"
-                          :take (taxy-make-take-function keys ement-room-list-keys))
-                         (taxy-fill room-session-vectors)
-                         (taxy-sort #'> #'item-latest-ts)
-                         (taxy-sort #'t<nil #'item-invited-p)
-                         (taxy-sort #'t<nil #'item-favourite-p)
-                         (taxy-sort #'t>nil #'item-low-priority-p)
-                         (taxy-sort #'t<nil #'item-unread-p)
-                         (taxy-sort #'t<nil #'item-space-p)
-                         ;; Within each taxy, left rooms should be sorted last so that one
-                         ;; can never be the first room in the taxy (unless it's the taxy
-                         ;; of left rooms), which would cause the taxy to be incorrectly
-                         ;; sorted last.
-                         (taxy-sort #'t>nil #'item-left-p)
-                         (taxy-sort* #'string< #'taxy-name)
-                         (taxy-sort* #'> #'taxy-latest-ts)
-                         (taxy-sort* #'t<nil (first-item item-unread-p))
-                         (taxy-sort* #'t<nil (first-item item-favourite-p))
-                         (taxy-sort* #'t<nil (first-item item-invited-p))
-                         (taxy-sort* #'t<nil (first-item item-buffer-p))
-                         (taxy-sort* #'t>nil (first-item item-space-p))
-                         (taxy-sort* #'t>nil (first-item item-low-priority-p))
-                         (taxy-sort* #'t>nil (first-item item-left-p)))))
-               (taxy-magit-section-insert-indent-items nil)
-               (inhibit-read-only t)
-               (format-cons (taxy-magit-section-format-items
-                             ement-room-list-columns ement-room-list-column-formatters taxy))
-               (pos (point))
-               (section-ident (when (magit-current-section)
-                                (magit-section-ident (magit-current-section)))))
-          (setf format-table (car format-cons)
-                column-sizes (cdr format-cons)
-                header-line-format (taxy-magit-section-format-header
-                                    column-sizes ement-room-list-column-formatters)
-                window-start (if (get-buffer-window buffer-name)
-                                 (window-start (get-buffer-window buffer-name))
-                               0))
-          (delete-all-overlays)
-          (erase-buffer)
-          (save-excursion
-            (taxy-magit-section-insert taxy :items 'first
-              ;; :blank-between-depth bufler-taxy-blank-between-depth
-              :initial-depth 0))
-          (goto-char pos)
-          (when (and section-ident (magit-get-section section-ident))
-            (goto-char (oref (magit-get-section section-ident) start)))))
+        (delete-all-overlays)
+        (erase-buffer)
+        (if (not room-session-vectors)
+            (insert "No joined rooms.  Use command `ement-join-room' to join a room, or `ement-directory' or `ement-directory-search' to find rooms.")
+          (let* ((taxy (cl-macrolet ((first-item
+                                      (pred) `(lambda (taxy)
+                                                (when (taxy-items taxy)
+                                                  (,pred (car (taxy-items taxy))))))
+                                     (name= (name) `(lambda (taxy)
+                                                      (equal ,name (taxy-name taxy)))))
+                         (thread-last
+                           (make-fn
+                            :name "Ement Rooms"
+                            :take (taxy-make-take-function keys ement-room-list-keys))
+                           (taxy-fill room-session-vectors)
+                           (taxy-sort #'> #'item-latest-ts)
+                           (taxy-sort #'t<nil #'item-invited-p)
+                           (taxy-sort #'t<nil #'item-favourite-p)
+                           (taxy-sort #'t>nil #'item-low-priority-p)
+                           (taxy-sort #'t<nil #'item-unread-p)
+                           (taxy-sort #'t<nil #'item-space-p)
+                           ;; Within each taxy, left rooms should be sorted last so that one
+                           ;; can never be the first room in the taxy (unless it's the taxy
+                           ;; of left rooms), which would cause the taxy to be incorrectly
+                           ;; sorted last.
+                           (taxy-sort #'t>nil #'item-left-p)
+                           (taxy-sort* #'string< #'taxy-name)
+                           (taxy-sort* #'> #'taxy-latest-ts)
+                           (taxy-sort* #'t<nil (name= "Buffers"))
+                           (taxy-sort* #'t<nil (first-item item-unread-p))
+                           (taxy-sort* #'t<nil (first-item item-favourite-p))
+                           (taxy-sort* #'t<nil (first-item item-invited-p))
+                           (taxy-sort* #'t>nil (first-item item-space-p))
+                           (taxy-sort* #'t>nil (name= "Low-priority"))
+                           (taxy-sort* #'t>nil (first-item item-left-p)))))
+                 (taxy-magit-section-insert-indent-items nil)
+                 (format-cons (taxy-magit-section-format-items
+                               ement-room-list-columns ement-room-list-column-formatters taxy))
+                 (pos (point))
+                 (section-ident (when (magit-current-section)
+                                  (magit-section-ident (magit-current-section)))))
+            (setf format-table (car format-cons)
+                  column-sizes (cdr format-cons)
+                  header-line-format (taxy-magit-section-format-header
+                                      column-sizes ement-room-list-column-formatters)
+                  window-start (if (get-buffer-window buffer-name)
+                                   (window-start (get-buffer-window buffer-name))
+                                 0))
+            (when ement-room-list-visibility-cache
+              (setf magit-section-visibility-cache ement-room-list-visibility-cache))
+            (add-hook 'kill-buffer-hook #'ement-room-list--cache-visibility nil 'local)
+            (save-excursion
+              (taxy-magit-section-insert taxy :items 'first
+                ;; :blank-between-depth bufler-taxy-blank-between-depth
+                :initial-depth 0))
+            (goto-char pos)
+            (when (and section-ident (magit-get-section section-ident))
+              (goto-char (oref (magit-get-section section-ident) start))))))
       (when display-buffer-action
-        (display-buffer buffer-name display-buffer-action))
+        (when-let ((window (display-buffer buffer-name display-buffer-action)))
+          (select-window window)))
       (when (get-buffer-window buffer-name)
         (set-window-start (get-buffer-window buffer-name) window-start))
       ;; NOTE: In order for `bookmark--jump-via' to work properly, the restored buffer
@@ -627,16 +704,26 @@ left."
                               (window-parameters
 			       (no-delete-other-windows . t))))))
 
-(defun ement-room-list-revert (_ignore-auto _noconfirm)
+(defun ement-room-list-revert (&optional _ignore-auto _noconfirm)
   "Revert current Ement-Room-List buffer."
   (interactive)
+  (with-current-buffer "*Ement Room List*"
+    ;; FIXME: This caching of the visibility only supports the main buffer with the
+    ;; default name, not any special ones with different names.
+    (setf ement-room-list-visibility-cache magit-section-visibility-cache))
   (ement-room-list :display-buffer-action nil))
 
-(defun ement-room-list-mouse-1 (event)
-  "Call `ement-room-list-RET' at EVENT."
-  (interactive "e")
-  (mouse-set-point event)
-  (call-interactively #'ement-room-list-RET))
+(defun ement-room-list-kill-buffer (room)
+  "Kill ROOM's buffer."
+  (interactive
+   (ement-with-room-and-session
+     (ignore ement-session)
+     (list ement-room)))
+  (pcase-let (((cl-struct ement-room (local (map buffer))) room)
+              (kill-buffer-query-functions))
+    (when (buffer-live-p buffer)
+      (kill-buffer buffer)
+      (ement-room-list-revert))))
 
 (declare-function ement-view-room "ement-room")
 (defun ement-room-list-RET ()
@@ -677,6 +764,15 @@ left."
 
 ;;;; Functions
 
+(defun ement-room-list--cache-visibility ()
+  "Save visibility cache.
+Sets `ement-room-list-visibility-cache' to the value of
+`magit-section-visibility-cache'.  To be called in
+`kill-buffer-hook'."
+  (ignore-errors
+    (when magit-section-visibility-cache
+      (setf ement-room-list-visibility-cache magit-section-visibility-cache))))
+
 ;;;###autoload
 (defun ement-room-list-auto-update (_session)
   "Automatically update the Taxy room list buffer.
